'''
For combining Trial microstates with Baseline microstates during Step-4: across-conditions
So the results from step-3 for Trial & Baseline are necessary input here

Prepared by Mengting Zhao, on Sep 15 2024
'''

from math_utilis import zero_mean, condensed_to_square
from utilis import read_subject_info, read_info, get_root_path, write_info
import numpy as np
from scipy import signal
import matplotlib.pyplot as plt
from scipy import stats
from scipy import spatial
import codecs, json
from collections import OrderedDict
import mne
from multiprocessing import Pool
import matplotlib.gridspec as gridspec
import itertools
from operator import itemgetter, attrgetter
from statsmodels.stats.multitest import multipletests
from sklearn import preprocessing
import os
from scipy.io import loadmat
from os.path import join as pjoin


def create_info():
    ## written by Mengting Zhao at Montreal, March11 2022
    biosemi_montage = mne.channels.make_standard_montage('biosemi64')
    ch = biosemi_montage.ch_names
    info = mne.create_info(ch_names=ch, sfreq=250.,ch_types='eeg')
    return info


class Microstate:
    def __init__(self, data):
        self.data = data.T
        self.data = zero_mean(self.data, 1)
        self.n_t = self.data.shape[0]
        self.n_ch = self.data.shape[1]
        self.gfp = None
        self.peaks = None

        self.cv = None
        self.maps = None
        self.label = None

        # All data
        self.label = None
        self.gev = None
        self.correlation = None
        self.duration = None
        self.coverage = None
        self.opt_k = -1
        self.opt_k_index = -1

        #All data from min_maps to max_maps
        self.cv_list = []
        self.gev_list = []
        self.maps_list = []
        self.label_list = []

    @staticmethod
    def orthogonal_dist(v, eeg_map):   # computes the main part of cost function Eq(2)
        return np.sum(v ** 2) - np.sum(np.sum((eeg_map * v), axis=1) ** 2)
        #
    @staticmethod
    def normalization(v, axis=1):
        return v / np.linalg.norm(v, axis=axis, keepdims=True)

    @staticmethod
    def global_dissimilarity(spatial_correlation):
        return np.sqrt(2 * (1 - spatial_correlation))

    def global_explained_variance(self, n_maps, correlation, label, data_std):
        gev = np.zeros(n_maps)
        for k in range(n_maps):
            gev[k] = np.sum(data_std[label == k] ** 2 * correlation[label == k, k]**2) / np.sum(data_std[label != -1]**2)
        return gev

    @staticmethod
    def global_explained_variance_sum(data, maps, distance=10, n_std=3, polarity=False):
        n_maps = maps.shape[0]
        n_ch = maps.shape[1]
        gev_peaks = np.zeros(n_maps)
        gev_raw = np.zeros(n_maps)
        gfp = data.std(axis=1)
        peaks, _ = signal.find_peaks(gfp, distance=distance, height=(gfp.mean() - n_std * gfp.std(), gfp.mean() + n_std * gfp.std()))
        correlation_peaks = Microstate.spatial_correlation(data[peaks], zero_mean(maps, 1).T, gfp[peaks], maps.std(axis=1), n_ch)
        correlation_raw = Microstate.spatial_correlation(data, zero_mean(maps, 1).T, gfp, maps.std(axis=1), n_ch)
        correlation_peaks = correlation_peaks if polarity else abs(correlation_peaks)
        correlation_raw = correlation_raw if polarity else abs(correlation_raw)
        label_peaks = np.argmax(correlation_peaks, axis=1)
        label_raw = np.argmax(correlation_raw, axis=1)
        for k in range(n_maps):
            gev_peaks[k] = np.sum(gfp[peaks][label_peaks == k] ** 2 * correlation_peaks[label_peaks == k, k]**2) / np.sum(gfp[peaks]**2)
            gev_raw[k] = np.sum(gfp[label_raw == k] ** 2 * correlation_raw[label_raw == k, k]**2) / np.sum(gfp**2)
        return np.sum(gev_peaks), np.sum(gev_raw)


    @staticmethod
    def opt_microstate_criteria(cvs):
        k_opt = np.argmin(cvs)
        return k_opt

    @staticmethod
    def max_evec(v, axis):
        data = np.dot(v.T, v)
        evals, evecs = np.linalg.eig(data)
        c = evecs[:, np.argmax(np.abs(evals))]
        c = np.real(c)
        return Microstate.normalization(c, axis)

    @staticmethod
    def spatial_correlation(averaged_v1, averaged_v2, std_v1, std_v2, n_ch):
        correlation = np.dot(averaged_v1, averaged_v2) / (n_ch * np.outer(std_v1, std_v2))
        return correlation

    @staticmethod
    def fit_back(data, maps, distance=10, n_std=3, polarity=False, instantaneous_eeg=False):
        if instantaneous_eeg:
            correlation = Microstate.spatial_correlation(data, zero_mean(maps, 1).T, data.std(axis=1),
                                                         maps.std(axis=1), data.shape[1])
            correlation = correlation if polarity else abs(correlation)
            label = np.argmax(correlation,axis=1)
            return label

        gfp = data.std(axis=1)
        peaks, _ = signal.find_peaks(gfp, distance=distance, height=(gfp.mean() - n_std * gfp.std(), gfp.mean() + n_std * gfp.std()))
        label = np.full(data.shape[0], -1)

        correlation = Microstate.spatial_correlation(data[peaks], zero_mean(maps, 1).T, gfp[peaks], maps.std(axis=1), data.shape[1])
        correlation = correlation if polarity else abs(correlation)
        label_peaks = np.argmax(correlation, axis=1)
        for i in range(len(peaks)):
            if i == 0:
                previous_middle = 0
                next_middle = int((peaks[i] + peaks[i + 1]) / 2)
            elif i == len(peaks) - 1:
                previous_middle = int((peaks[i] + peaks[i - 1]) / 2)
                next_middle = len(label)
            else:
                previous_middle = int((peaks[i] + peaks[i - 1]) / 2)
                next_middle = int((peaks[i] + peaks[i + 1]) / 2)
            label[previous_middle:next_middle] = label_peaks[i]
        return label

    @staticmethod
    def microstates_duration(label, n_maps):
        duration = [[] for i in range(n_maps)]
        j = label[0]
        count = 1
        for i in range(1, len(label)):
            if j != label[i]:
                duration[j].append(count)
                j = label[i]
                count = 1
            elif j == label[i]:
                count += 1
        for i, temp in enumerate(duration):
            if i == label[0]:
                duration[i] = np.asarray(temp[1::]).mean() if len(temp[1::]) > 0 else 0
            elif i == label[-1]:
                duration[i] = np.asarray(temp[0:-1]).mean() if len(temp[0:-1]) > 0 else 0
            else:
                duration[i] = np.asarray(temp).mean() if len(temp) > 0 else 0

        return duration

    @staticmethod
    def microstates_occurrence(label, n_maps, sfreq):
        occurence = [[] for i in range(n_maps)]
        for i in range(0, len(label), sfreq):
            temp = np.asarray(label[i:i + sfreq])
            for j in range(n_maps):
                occurence[j].append(np.argwhere(temp == j).shape[0])
        for i, temp in enumerate(occurence):
            occurence[i] = np.mean(temp) if len(temp) > 0 else 0
        return occurence

    @staticmethod
    def microstates_coverage(label, n_maps):
        coverage = []
        n_label = len(label)
        for i in range(n_maps):
            coverage.append(np.argwhere(label == i).shape[0] / n_label)
        return coverage

    @staticmethod
    def microstates_parameters(data, maps, distance=10, n_std=3, polarity=False, sfreq=500, epoch=2):

        n_maps = len(maps)
        res = {'duration': [], 'occurrence': [], 'coverage': []}
        data = zero_mean(data.T, 1)

        for i in range(0, data.shape[0], sfreq * epoch):
            data_epoch = data[i: i + sfreq * epoch]
            label = Microstate.fit_back(data_epoch, maps, distance, n_std, polarity)  # fit-back is applied here

            res['duration'].append(Microstate.microstates_duration(label, n_maps))
            res['occurrence'].append(Microstate.microstates_occurrence(label, n_maps, sfreq))
            res['coverage'].append(Microstate.microstates_coverage(label, n_maps))
        return res

    def cross_validation(self, var, n_maps):
        return var * (self.n_ch - 1) ** 2 / (self.n_ch - n_maps - 1.) ** 2

    def variance(self, label, maps, data):
        var = np.sum(data ** 2) - np.sum(np.sum(maps[label, :] * data, axis=1) ** 2)
        var /= (self.n_t * (self.n_ch-1))
        return var

    def gfp_peaks(self, distance=10, n_std=3):  # first step as mentioned in the paper
        gfp = self.data.std(axis=1)
        peaks, _ = signal.find_peaks(gfp, distance=distance, height=(gfp.mean() - n_std * gfp.std(), gfp.mean() + n_std * gfp.std()))
        self.peaks = peaks
        self.gfp = gfp

    def kmeans_modified(self, data, data_std, n_runs=10, n_maps=4, maxerr=1e-6, maxiter=1000, polarity=False):
        n_gfp = data_std.shape[0]
        cv_list = []
        gev_list = []
        maps_list = []
        label_list = []
        for run in range(n_runs):
            rndi = np.random.permutation(n_gfp)[:n_maps]
            maps = Microstate.normalization(data[rndi, :], axis=1)
            n_iter = 0
            var0 = 1.0
            var1 = 0.0
            while ((np.abs(var0 - var1) / var0 > maxerr) & (n_iter < maxiter)):
                label = np.argmax(np.dot(data, maps.T) ** 2, axis=1)
                for k in range(n_maps):
                    data_k = data[label == k, :]
                    maps[k, :] = Microstate.max_evec(data_k, 0)
                var1 = var0
                var0 = Microstate.orthogonal_dist(data, maps[label, :])
                var0 /= (n_gfp * (self.n_ch - 1))

            label, correlation, cv, gev = self.optimize_k(maps=maps, data=data, data_std=data_std, polarity=polarity)
            cv_list.append(cv)
            gev_list.append(gev)
            maps_list.append(maps)
            label_list.append(label)
            opt = Microstate.opt_microstate_criteria(cv_list)
        return cv_list[opt], gev_list[opt], maps_list[opt], label_list[opt]


    def opt_microstate(self, min_maps=2, max_maps=10, distance=100, n_std=3, n_runs=100, maxerr=1e-6, maxiter=1000, polarity=False, peaks_only=True, method='kmeans_modified'):
        # the K in Kmeans-clustering changes from 2 to 10 as min_maps=2 and max_maps = 10
        self.gfp_peaks(distance=distance, n_std=n_std)
        if peaks_only:  # send the EEG data at peaks to modified-Kmeans
            temp_data = self.data[self.peaks]
            temp_data_std = self.gfp[self.peaks]
            temp_max_maps = min(temp_data.shape[0], max_maps)
        else:
            temp_data = self.data
            temp_data_std = self.gfp
            temp_max_maps = min(temp_data.shape[0], max_maps)

        if method == 'kmeans_modified':
            for n_maps in range(min_maps, temp_max_maps+1):
                print("kmeans_number:{number}".format(number=n_maps))
                cv, gev, maps, label = self.kmeans_modified(data=temp_data, data_std=temp_data_std, n_runs=n_runs, n_maps=n_maps, maxerr=maxerr, maxiter=maxiter, polarity=polarity)
                self.cv_list.append(cv)
                self.gev_list.append(gev)
                self.maps_list.append(maps)
                self.label_list.append(label)

        elif method == 'aahc':
            self.cv_list, self.gev_list, self.maps_list, self.label_list = self.aahc(data=temp_data, data_std=temp_data_std, min_maps=min_maps, max_maps=temp_max_maps, polarity=polarity)

        self.opt_k_index = Microstate.opt_microstate_criteria(self.cv_list)

        self.cv = self.cv_list[self.opt_k_index]
        self.gev = self.gev_list[self.opt_k_index]
        self.maps = self.maps_list[self.opt_k_index]
        self.label = self.label_list[self.opt_k_index]

        self.opt_k = len(self.gev)

        self.gev_list = [temp.tolist() for temp in self.gev_list]
        self.maps_list = [temp.tolist() for temp in self.maps_list]
        self.label_list = [temp.tolist() for temp in self.label_list]

    def optimize_k(self, maps, data=None, data_std=None, polarity=False):
        # this is used in kmeans_modified() function
        if data is None and data_std is None:
            data = self.data
            data_std = self.gfp
        n_maps = len(maps)
        correlation = Microstate.spatial_correlation(data, zero_mean(maps, 1).T, data_std, maps.std(axis=1),self.n_ch)
        correlation = correlation if polarity else abs(correlation)
        label = np.argmax(correlation, axis=1)
        var = self.variance(label=label, maps=maps, data=data)
        cv = self.cross_validation(var, n_maps)
        gev = self.global_explained_variance(n_maps=n_maps, correlation=correlation, label=label, data_std=data_std)
        return label, correlation, cv, gev


    def aahc(self, data, data_std, min_maps, max_maps, polarity=False):
        maps = data
        n_maps = len(maps)
        label_list = [[k] for k in range(len(maps))]
        cv_list = []
        gev_list = []
        maps_list = []
        res_label_list = []
        while n_maps > (min_maps - 1):
            print("n_maps:%d" % n_maps)
            correlation = Microstate.spatial_correlation(data, zero_mean(maps, 1).T, data_std, maps.std(axis=1),self.n_ch)
            correlation = correlation if polarity else abs(correlation)
            label = np.argmax(correlation, axis=1)
            gev = self.global_explained_variance(n_maps, correlation, label, data_std)
            if max_maps >= n_maps >= min_maps:
                var = self.variance(label, maps, data=data)
                cv = self.cross_validation(var, n_maps)
                cv_list.append(cv)
                maps_list.append(maps)
                gev_list.append(gev)
                res_label_list.append(label)
                if n_maps == min_maps:
                    break
            excluded_k = np.argmin(gev)
            maps = np.vstack((maps[:excluded_k, :], maps[excluded_k+1:, :]))
            re_label = label_list.pop(excluded_k)
            re_cluster = []
            for k in re_label:
                correlation = Microstate.spatial_correlation(data[k, :], zero_mean(maps, 1).T, data_std[k], maps.std(axis=1),self.n_ch)
                correlation = correlation if polarity else abs(correlation)
                new_label = np.argmax(correlation)
                re_cluster.append(new_label)
                label_list[new_label].append(k)

            for i in re_cluster:
                idx = label_list[i]
                maps[i] = Microstate.max_evec(data[idx, :], 0)
            n_maps = len(maps)

        return cv_list[::-1], gev_list[::-1], maps_list[::-1], res_label_list[::-1]


class MeanMicrostate:
    def __init__(self, data, n_k, n_ch, n_condition):
        self.data = data
        self.n_k = n_k
        self.n_ch = n_ch
        self.n_condition = n_condition
        data_concatenate = np.zeros((1, n_ch))
        for i in range(n_condition):
            data_concatenate = np.concatenate((data_concatenate, data[i]), axis=0)  # axis=0 from Wenjun's algo

        self.data_concatenate = data_concatenate[1::, :]


    def label_two_microstates(self, microstates, mean_microstates, polarity=False):
        similarity_matrix = np.zeros((self.n_k, self.n_k))
        for i in range(self.n_k):
            for j in range(self.n_k):
                similarity_matrix[i, j] = stats.pearsonr(microstates[i], mean_microstates[j])[0]

        comb = [zip(perm, [i for i in range(self.n_k)]) for perm in itertools.permutations([j for j in range(self.n_k)], self.n_k)]
        res = []
        for i, item in enumerate(comb):
            s = 0
            comb_list = []
            sign = []
            for item_j in item:
                comb_list.append(item_j)
                similarity = similarity_matrix[item_j[0], item_j[1]]
                if similarity < 0:
                    sign.append(-1)
                else:
                    sign.append(1)
                s = s + similarity if polarity else s + abs(similarity)
            res.append((s/len(comb_list), comb_list, sign))
        sorted_res = sorted(res, key=itemgetter(0), reverse=True)
        return sorted_res[0]

    def label_microstates(self, mul_microstates, mean_microstates, polarity=False):
        label = []
        sign = []
        similarity = []
        for microstates in mul_microstates:
            s = self.label_two_microstates(microstates, mean_microstates, polarity)
            for index, item in enumerate(s[1]):
                label.append(item[0])
            sign.extend(s[2])
            similarity.append(s[0])
        return label, sign, np.mean(similarity), np.std(similarity)

    def reorder_microstates(self, mul_microstates, mean_microstates, polarity=False):
        res = []
        for microstates in mul_microstates:
            s = self.label_two_microstates(microstates, mean_microstates, polarity)
            sorted_index = [i[0] for i in s[1]]
            sign = np.repeat(s[2], self.n_ch).reshape(-1,self.n_ch)
            microstate_updated = np.asarray(microstates)[sorted_index] * sign
            res.append(microstate_updated.tolist())

        return res

    def update_mean_microstates(self, label, sign, polarity=False):
        label = np.asarray(label)
        mean_microsate_updated = np.zeros((self.n_k, self.n_ch))
        for i in range(self.n_k):
            index = np.argwhere(label == i).reshape(-1)
            maps = self.data_concatenate[index, :]
            # if not polarity:
            temp = np.asarray(sign)[index].reshape(self.n_condition, 1)
            temp = np.repeat(temp, self.n_ch, axis=1)
            maps = maps * temp
            mean_microsate_updated[i, :] = np.mean(maps, axis=0)
        return mean_microsate_updated

    ####### tested in 'main' function with success ##########
    def mean_microstates(self, n_runs=100, maxiter=100):
        n_data_concatenate = len(self.data_concatenate)
        maps_list = []
        label_list = []
        mean_similarity_list = []
        std_similarity_list = []
        for run in range(n_runs):
            print(run)
            mean_similarity_run = []
            std_similarity_run = []
            label_run = []
            maps_run = []
            rndi = np.random.permutation(n_data_concatenate)[:self.n_k]
            # maps = Microstate.normalization(self.data_concatenate[rndi, :], axis=1)

            maps = self.data_concatenate[rndi, :]
            label, sign, mean_similarity, std_similarity = self.label_microstates(self.data, maps, False)
            iter_num = 0
            while iter_num < maxiter:
                iter_num += 1
                maps_updated = self.update_mean_microstates(label, sign, False)
                label_updated, sign, mean_similarity, std_similarity = self.label_microstates(self.data, maps_updated, False)
                if label == label_updated:
                    maps_list.append(maps_updated)
                    label_list.append(label)
                    mean_similarity_list.append(mean_similarity)
                    std_similarity_list.append(std_similarity)
                    break
                else:
                    mean_similarity_run.append(mean_similarity)
                    std_similarity_run.append(std_similarity)
                    label_run.append(label_updated)
                    maps_run.append(maps_updated)
                    label = label_updated
            else:
                index = np.argmax(np.asarray(mean_similarity_run))
                mean_similarity_list.append(mean_similarity_run[index])
                std_similarity_list.append(std_similarity_run[index])
                label_list.append(label_run[index])
                maps_list.append(maps_run[index])

        index = np.argmax(mean_similarity_list)
        return maps_list[index], label_list[index], mean_similarity_list[index], std_similarity_list[index]


def batch_microstate_parameters(para):
    data = para[0]
    maps = para[1]
    distance = para[2]
    n_std = para[3]
    polarity = para[4]
    sfreq = para[5]
    epoch = para[6]
    _ = Microstate(np.asarray(data))
    return Microstate.microstates_parameters(np.asarray(data), np.asarray(maps), distance, n_std, polarity, sfreq, epoch)

def batch_microstate_state(data, topographies, task_name):
    for a_task_name in task_name:
        microstate = Microstate(data[a_task_name])
        microstate.gfp_peaks()
        microstate.maps = np.asarray(topographies[a_task_name]["maps_list"][topographies[a_task_name]["opt_k_index"]])
        microstate.microstate_state()

def batch_order_mean_microstate(para):
    data = para[0]
    n_k = para[1]
    n_ch = para[2]
    n_condition = para[3]
    mean_microstate = para[4]
    microstate = MeanMicrostate(data, n_k, n_ch, n_condition)
    res = microstate.reorder_microstates(microstate.data, mean_microstate, polarity=False)
    return res

# updated on March 13 2022 by Mengting Zhao for CTA analysis
def batch_mean_microstate(para):
    # batch_mean_microstate(maps, n_k, n_ch, nb_trials)
    data = para[0]
    n_k = para[1]
    n_ch = para[2]
    n_condition = para[3]
    microstates = MeanMicrostate(data, n_k, n_ch, n_condition)
    eegmaps, label, mean_similarity, std_similarity = microstates.mean_microstates()
    return {"maps": eegmaps, "label": label, "mean_similarity": mean_similarity, "std_similarity": std_similarity}


def batch_microstate(para):
    data = para[0]
    peaks_only = para[1]
    min_maps = para[2]
    max_maps = para[3]
    microstate = Microstate(data)
    microstate.opt_microstate(min_maps, max_maps, n_std=3, n_runs=100, peaks_only=peaks_only, method='kmeans_modified')
    return microstate


def filter_data(data):
    return mne.filter.filter_data(data, 500, 1., 30.)


def batch_filter(clean_data_fname, subjects, data_fname, data_fname_save_dict, data_fname_save, task_name):
    for subject in subjects:
        print(subject)
        res = OrderedDict()
        data_text = codecs.open(clean_data_fname + "\\" + subject + data_fname, 'r', encoding='utf-8').read()
        data = json.loads(data_text)
        pool = Pool(8)
        multi_res = [pool.apply_async(filter_data, (data[a_task_name]['task_data'],)) for a_task_name in task_name]
        pool.close()
        pool.join()
        for i in range(len(task_name)):
            res[task_name[i]] = multi_res[i].get().tolist()
        json.dump(res, codecs.open(data_fname_save_dict + "\\" + subject + data_fname_save, 'w', encoding='utf-8'), separators=(',', ':'), sort_keys=True, indent=4)

### updated by Mengting Zhao on March 17 for CTA analysis
# ###### to be tested in main() function, not working ##############
# def eegmaps_parameters(clean_data_fname, subjects, data_fname, eegmaps_fname, clean_data_fname_save, data_fname_save, task_name):
#     # # this version takes the computed microstate.json under each subject as input
#     ### eegmaps_name should be the full name of the saving file for globel eegmaps
#     text_eegmaps = codecs.open(eegmaps_fname, 'r', encoding='utf-8').read()
#     eegmaps = json.loads(text_eegmaps)
#     eegmaps = eegmaps['maps'] if 'maps' in eegmaps else eegmaps
#     for subject in subjects:
#         print(subject)
#         temp_res = []
#         res = OrderedDict()
#         # pool = Pool(len(task_name))
#         #### if clean_data_fname = 'F:\FlightSim_experiment_2021\Concordia_preprocessed\microstates_subjects'
#         #### and data_fname = '_microstate_labels'
#         data_text = codecs.open(clean_data_fname + "\\" + subject + "\\" + 'microstates' + "\\" + subject + data_fname, 'r', encoding='utf-8').read()
#         data = json.loads(data_text)
#         for a_task_name in task_name:
#             # # temp_res.append(pool.apply_async(batch_microstate_parameters, ([data[a_task_name], eegmaps, 10, 3, False, 500, 2],)))
#             # # _ = Microstate(np.asarray(data[a_task_name]))
#             # vvv = np.asarray(data[a_task_name]).reshape((-1, 1))
#             # _ = Microstate(np.asarray(data[a_task_name]))
#             _ = Microstate.microstates_parameters(np.asarray(data[a_task_name]), np.asarray(eegmaps), 10, 3, False, 500, 2)
#             temp_res.append(Microstate.microstates_parameters(np.asarray(data[a_task_name]), np.asarray(eegmaps), 10, 3, False, 500, 2))
#         # pool.close()
#         # pool.join()
#         for i, a_task_name in enumerate(task_name):
#             res[a_task_name] = temp_res[i].get()
#         json.dump(res, codecs.open(clean_data_fname_save + "\\" + subject + data_fname_save, 'w', encoding='utf-8'), separators=(',', ':'), sort_keys=True, indent=4)


### updated by Mengting Zhao on March 17 for CTA analysis
######  tested with success ###############################
def eegmaps_parameters(data_dir, subjects, data_fname_save, eegmaps_fname, task_name):
    ### eegmaps_name should be the full name of the saving file for globel eegmaps
    text_eegmaps = codecs.open(eegmaps_fname, 'r', encoding='utf-8').read()
    eegmaps = json.loads(text_eegmaps)
    eegmaps = eegmaps['maps'] if 'maps' in eegmaps else eegmaps
    for subject in subjects:
        print(subject)
        # data = load_data(clean_data_fname + "\\" + subject + data_fname)  #from Wenjun's code
        subject_path = pjoin(data_dir, subject)
        # data_fname_save_dict = pjoin(subject_path, r'microstates')
        res = OrderedDict()
        temp_res = []
        for q in range(nb_trials):
            min_maps_lis = [1 for i in range(len(path_name))]
            max_maps_lis = [10 for i in range(len(path_name))]
            trial_path = subject_path + "\\" + r'clean_data_matlab' + "\\" + r'data' + "\\" + path_name[q]
            set_name = pjoin(trial_path, lis_name[q])
            EEG_epochs= mne.io.read_epochs_eeglab(set_name)

            # deal with the channel naming issue when adding .loc information
            mapping = {'Afz': 'AFz'}
            EEG_epochs.rename_channels(mapping, allow_duplicates=True, verbose=None)
            names_channel = EEG_epochs.ch_names
            n_chs = len(names_channel)

            EEG_epochs = EEG_epochs.set_montage('biosemi64')
            # # get the IDs of the events in the dataset
            # ids_event = EEG_epochs.event_id
            nb_epochs = EEG_epochs.selection.size
            n_epo[q] = nb_epochs

            data_test = EEG_epochs.get_data()
            data_res = []
            for i in range(nb_epochs):
                temp_epoch = EEG_epochs[i]
                temp_evoked = temp_epoch.average()
                temp_key = list(temp_epoch.event_id)[0]
                temp_data = temp_epoch.get_data()[0]
                if i ==0:
                    data_res = temp_data
                else:
                    data_res = np.concatenate((data_res,temp_data), axis=1)

            temp_res.append(Microstate.microstates_parameters(np.asarray(data_res), np.asarray(eegmaps), 10, 3, False, 250, 2))

        # pool.join()
        for i, a_task_name in enumerate(task_name):
            # res[a_task_name] = temp_res[i].get()
            res[a_task_name] = temp_res[i]
        json.dump(res, codecs.open(subject_path + "\\" + r'microstates' + "\\" + subject + data_fname_save, 'w', encoding='utf-8'), separators=(',', ':'), sort_keys=True, indent=4)


## updated by Mengting on March 18 2022 for CTA analysis
def eegmaps_parameters_across_runs(data_dir, subjects, data_fname, data_fname_save, condition_name, tasks_name):
    ## the condition swich part is added by Mengting
    conditions = {}
    # res = OrderedDict()
    for condition in condition_name:
        conditions[condition] = []
        for task in tasks_name:
            if task.startswith(condition):
                conditions[condition].append(task)

    for subject in subjects:
        print(subject)
        # data = load_data(clean_data_fname + "\\" + subject + data_fname)
        clean_data_fname = data_dir + "\\" + subject + "\\" + 'microstates'
        data_text = codecs.open(clean_data_fname + "\\" + subject + data_fname, 'r', encoding='utf-8').read()
        data = json.loads(data_text)
        res = OrderedDict()
        for condition in condition_name:
            # maps = [data[a_task_name]['maps_list'][n_k_index] for a_task_name in conditions[condition]]
            # # condition_res.append(pool.apply_async(batch_mean_microstate, ([maps, n_k, n_chs, len(conditions[condition])],)))
            # vvv = MeanMicrostate(maps, n_k, n_chs, len(conditions[condition]))
            duration = []
            coverage = []
            occurrence = []
            res[condition] = {}
            for a_task_name in conditions[condition]:
                duration.extend(data[a_task_name]["duration"])
                coverage.extend(data[a_task_name]["coverage"])
                occurrence.extend(data[a_task_name]["occurrence"])
            res[condition]['duration'] = exclude_zero_mean(np.asarray(duration)).tolist()
            res[condition]['coverage'] = np.asarray(coverage).mean(axis=0).tolist()
            res[condition]['occurrence'] = exclude_zero_mean(np.asarray(occurrence)).tolist()
        json.dump(res, codecs.open(clean_data_fname + "\\" + subject + data_fname_save, 'w', encoding='utf-8'), separators=(',', ':'), sort_keys=True, indent=4)


def exclude_zero_mean(data):
    sum = np.sum(data, axis=0)
    col = (data != 0).sum(0)
    return sum/col


################# updated by Mengting Zhao on March14 2022 for CTA analysis for pilot training research#####################
### function ready for use on CTA data
########## tested with success ###############
def eegmaps_across_subjects(clean_data_dict, subjects, data_fname, data_fname_save_dict, data_fname_save, condition_name, n_k, n_chs):
    res = OrderedDict()
    condition_res = []
    n_condition = len(condition_name)
    # pool = Pool(n_condition)
    for a_condition_name in condition_name:
        maps = []
        for subject in subjects:
            print(subject)
            # data_text = codecs.open(clean_data_dict + "\\" + subject + "\\" + 'microstates'+"\\" + subject + data_fname, 'r', encoding='utf-8').read()
            data_text = codecs.open(clean_data_dict + "\\" + subject + data_fname, 'r', encoding='utf-8').read()
            data = json.loads(data_text)
            maps.append(data[a_condition_name]['maps'])

        # condition_res.append(pool.apply_async(batch_mean_microstate, ([maps, n_k, n_ch, len(subjects)],)))
        vvv = MeanMicrostate(maps, n_k, n_chs, len(subjects))
        eegmaps, label, mean_similarity, std_similarity = vvv.mean_microstates()
        condition_res.append({"maps": eegmaps, "label": label, "mean_similarity": mean_similarity, "std_similarity": std_similarity})
    for i, a_condition_name in enumerate(condition_name):
        temp = condition_res[i]
        res[a_condition_name] = {'maps': temp["maps"].tolist(), 'label': temp["label"], 'mean_similarity': temp['mean_similarity'], 'std_similarity': temp['std_similarity']}
    json.dump(res, codecs.open(data_fname_save_dict + "\\" + data_fname_save, 'w', encoding='utf-8'), separators=(',', ':'), sort_keys=True, indent=4)


#### updated by Mengting on April 17 for computation on Speed Server at Concordia
def eegmpas_across_subjects_linux(clean_data_fname, subjects, data_fname, data_fname_save_dict, data_fname_save, condition_name, n_k, n_ch):
    res = OrderedDict()
    condition_res = []
    n_condition = len(condition_name)
    pool = Pool(n_condition)
    for a_condition_name in condition_name:
        maps = []
        for subject in subjects:
            data = load_data(clean_data_fname + "/" + subject + data_fname)
            maps.append(data[a_condition_name]['maps'])
        condition_res.append(pool.apply_async(batch_mean_microstate, ([maps, n_k, n_ch, len(subjects)],)))
    for i, a_condition_name in enumerate(condition_name):
        temp = condition_res[i].get()
        res[a_condition_name] = {'maps': temp["maps"].tolist(), 'label': temp["label"], 'mean_similarity': temp['mean_similarity'], 'std_similarity': temp['std_similarity']}
    json.dump(res, codecs.open(data_fname_save_dict + "/" + data_fname_save, 'w', encoding='utf-8'), separators=(',', ':'), sort_keys=True, indent=4)


################# updated by Mengting Zhao on April 25 2022 #####################
def two_eegmaps_across_conditions(clean_data_dict_1, data_fname_1, clean_data_dict_2, data_fname_2, data_fname_save_dict, data_fname_save, condition_name, n_k, n_chs):
    maps = []
    a_name = 'rest'
    ## load the information for 'rest' from the Trial microstates
    data_text = codecs.open(clean_data_dict_1 + "\\" + data_fname_1, 'r', encoding='utf-8').read()
    data = json.loads(data_text)
    maps.append(data[a_name]['maps'])

    for a_condition_name in condition_name:
        # data = load_data(clean_data_dict + "\\" + data_fname)
        # maps.append(data[a_condition_name]['maps'])
        data_text = codecs.open(clean_data_dict_1 + "\\" + data_fname_1, 'r', encoding='utf-8').read()
        data = json.loads(data_text)
        maps.append(data[a_condition_name]['maps'])
    for a_condition_name in condition_name:
        # data = load_data(clean_data_dict + "\\" + data_fname)
        # maps.append(data[a_condition_name]['maps'])
        data_text = codecs.open(clean_data_dict_2 + "\\" + data_fname_2, 'r', encoding='utf-8').read()
        data = json.loads(data_text)
        maps.append(data[a_condition_name]['maps'])
    microstate = MeanMicrostate(maps, n_k, n_chs, 2*len(condition_name)+1)
    temp = microstate.mean_microstates()
    res = {'maps': temp[0].tolist(), 'label': temp[1], 'mean_similarity': temp[2], 'std_similarity': temp[3]}
    json.dump(res, codecs.open(data_fname_save_dict + "\\" + data_fname_save, 'w', encoding='utf-8'), separators=(',', ':'), sort_keys=True, indent=4)

## updated by Mengting on April 17
def eegmaps_across_conditions_linux(clean_data_fname, data_fname, data_fname_save_dict, data_fname_save, condition_name, n_k, n_ch):
    ## Pool is not used in this function
    maps = []
    for a_condition_name in condition_name:
        data = load_data(clean_data_fname + "/" + data_fname)
        maps.append(data[a_condition_name]['maps'])
    microstate = MeanMicrostate(maps, n_k, n_ch, len(condition_name))
    temp = microstate.mean_microstates()
    res = {'maps': temp[0].tolist(), 'label': temp[1], 'mean_similarity': temp[2], 'std_similarity': temp[3]}
    json.dump(res, codecs.open(data_fname_save_dict + "/" + data_fname_save, 'w', encoding='utf-8'), separators=(',', ':'), sort_keys=True, indent=4)



def plot_eeg(data, condition_name, n_k):
    info = create_info()
    for k in range(n_k):
        ax_all = gridspec.GridSpec(10, 5, figure=plt.figure(figsize=(10, 100)))
        for a_condition_name in condition_name:
            maps = data[a_condition_name]
            for row in range(10):
                for col in range(5):
                    ax = plt.subplot(ax_all[row, col])
                    mne.viz.plot_topomap(np.asarray(maps[row][k]), info, show=False, axes=ax, image_interp='spline36', contours=6)
        plt.show()





def plot_eegmaps(maps, task_name):
    info = create_info()
    min_value = maps[0].min()
    max_value = maps[0].max()
    if len(task_name) == 1:
        ax_all = gridspec.GridSpec(1, len(task_name), figure=plt.figure(figsize=(10, 100)))
        for j in range(len(maps[0])):
            ax = plt.subplot(ax_all[j])
            mne.viz.plot_topomap(maps[0].T[:, j], info, show=False, axes=ax, image_interp='spline36', contours=6, vmin=min_value, vmax=max_value)
        plt.show()
    else:
        similarity = []
        row = len(task_name)
        col = 0
        for i in range(1, len(task_name)):
            temp = eegmaps_similarity(maps[0], maps[i])
            similarity.append(temp)
            col = max(col, max([item['index'] for item in [*temp.values()]])+1)
        ax_all = gridspec.GridSpec(row, col, figure=plt.figure(figsize=(10, 100)))
        ax_all.update(hspace=0.5)
        for i in range(row):
            for j in range(len(maps[i])):
                if i == 0:
                    ax = plt.subplot(ax_all[i, j])
                else:
                    ax = plt.subplot(ax_all[i, similarity[i-1][j]['index']])
                    ax.set_title(round(similarity[i-1][j]['similarity'], 4), fontsize=5)
                mne.viz.plot_topomap(maps[i].T[:, j], info, show=False, axes=ax, image_interp='spline36', contours=6, vmin=min_value, vmax=max_value)
        plt.show()


def eegmaps_similarity_across_runs(clean_data_fname, subjects, data_fname, task_name, opt_k_index):
    res = []
    for subject in subjects:
        data_text = codecs.open(clean_data_fname + "\\" + subject + data_fname, 'r', encoding='utf-8').read()
        data = json.loads(data_text)
        res_temp = []
        for a_task_name in task_name:
            for comb in itertools.combinations([i for i in range(1, len(task_name)+1)], 2):
                similarity = eegmaps_similarity(np.asarray(data[str(comb[0]) + "_" + a_task_name]["maps_list"][opt_k_index]), np.asarray(data[str(comb[1]) + "_" + a_task_name]["maps_list"][opt_k_index]), 0.0)
                averaged_similarity = eegmaps_averaged_similarity(similarity)
                res_temp.append(averaged_similarity)
        res.append(res_temp)
    return res



def eegmaps_similarity_highest_comb(data):
    data = sorted(data, key=itemgetter(1), reverse=True)
    res = []
    for item in data:
        for item_res in res:
            if len(set(item[0]).intersection(set(item_res[0]))) != 0:
                break
        else:
            res.append(item)
    return res

def eegmaps_similarity_highest_merged(comb, eegmaps, n_runs, opt_k):
    res_eegmaps = []
    for item in comb:
        res = []
        for i in range(len(item[0])):
            temp = 0
            for j in range(len(item[0])):
                temp += abs(stats.pearsonr(eegmaps[i][item[0][i] % n_runs], eegmaps[j][item[0][j] % n_runs])[0])
            res.append(temp/len(item[0]))
        index = np.argmax(np.asarray(res))
        res_eegmaps.append(eegmaps[task_index_runs(item[0][index], opt_k)-1][item[0][index]%opt_k].tolist())
    return res_eegmaps


def task_index_runs(index, opt_k):
    if index < opt_k:
        return 1
    elif opt_k <= index < 2*opt_k:
        return 2
    elif 2*opt_k <= index < 3*opt_k:
        return 3

def eegmaps_averaged_similarity(similarity):
    res = 0
    for key, value in similarity.items():
        res += value['similarity']
    return res / len(similarity)


def eegmaps_similarity(baseline, task, threshold=0.0):
    res = {}
    similarity = np.zeros((baseline.shape[0], task.shape[0]))
    for i in range(baseline.shape[0]):
        for j in range(task.shape[0]):
            similarity[i][j] = abs(stats.pearsonr(baseline[i, :], task[j, :])[0])
    iter = min(baseline.shape[0], task.shape[0])
    while iter != 0:
        max_similarity = np.max(similarity)
        if max_similarity < threshold:
            iter -= 1
            continue
        temp = np.argwhere(similarity == max_similarity)
        row = temp[0][0]
        col = temp[0][1]
        res[col] = {'index': row, 'similarity': max_similarity}
        iter -= 1
        similarity[row, ::] = -3
        similarity[::, col] = -3
    if len(res) != task.shape[0]:
        task_index = [*res]
        diff = list(set([i for i in range(task.shape[0])]) - set(task_index))
        temp = baseline.shape[0]
        for i in diff:
            res[i] = {'index': temp, 'similarity': -3}
            temp += 1
    return res


def concatenate_data_by_condition(data, task_condition, exclude_task):
    res = OrderedDict()
    for condition in task_condition:
        for task_name, task_data in data.items():
            if task_name not in exclude_task:
                if condition == task_name.split("_")[1]:
                    res[condition] = np.asarray(task_data) if condition not in res else np.concatenate((res[condition], np.asarray(task_data)), axis=1)
    return res


#######the content has been tested in main( ) with success ##########
def opt_microstate_tasks(data, task_name, min_maps, max_maps, peaks_only=True):
    res = OrderedDict()
    pool = Pool(11)
    multi_res = []
    for i in range(len(task_name)):
        multi_res.append(pool.apply_async(batch_microstate, ([np.asarray(data[task_name[i]]), peaks_only, min_maps[i], max_maps[i]],)))
    pool.close()
    pool.join()
    for i in range(len(task_name)):
        temp = multi_res[i].get()
        res[task_name[i]] = {'cv_list': temp.cv_list, 'gev_list': temp.gev_list, 'maps_list': temp.maps_list,
                             'opt_k': temp.opt_k, 'opt_k_index': int(temp.opt_k_index), 'min_maps': min_maps[i],
                             'max_maps': max_maps[i]}
    return res

######## updated by Mengting on March 16 2022 for CTA analysis
#### to be tested .........
def opt_microstate_across_subjects(clean_data_fname, subjects, data_fname, data_fname_save, task_name, peaks_only=False):
    ### clean_data_fname should be data_dir containing the statistical data folder and also individual's result folders
    # mean_opt_k = microstate_stat(clean_data_fname + "\\" + "_1_30_microstate_stat.json")  # may cause error
    mean_opt_k = read_microstate_stat(clean_data_fname + "\\" + "microstates_stats" + "\\" + "_microstate_stat.json")
    task_data = OrderedDict()
    for a_task_name in task_name:
        for subject in subjects:
            data_text = codecs.open(clean_data_fname + "\\" + subject + "\\" + 'microstates' + "\\" + subject + data_fname, 'r', encoding='utf-8').read()
            data = json.loads(data_text)
            opt_k_index = data[a_task_name]["opt_k_index"]
            temp = np.asarray(data[a_task_name]["maps_list"][opt_k_index]).T
            task_data[a_task_name] = temp if a_task_name not in task_data else np.concatenate((task_data[a_task_name], temp), axis=1)
    # res = opt_microstate_tasks(task_data, task_name, [1 for i in range(len(task_name))], [10 for i in range(len(task_name))], peaks_only)

    res = opt_microstate_tasks(task_data, task_name, [mean_opt_k[a_task_name]["mean_opt_k"] for a_task_name in task_name], [mean_opt_k[a_task_name]["mean_opt_k"] for a_task_name in task_name], peaks_only)
    json.dump(res, codecs.open(clean_data_fname +  "\\" + "microstates_stats" + "\\" + data_fname_save, 'w', encoding='utf-8'), separators=(',', ':'), sort_keys=True, indent=4)


def opt_microstate_subjects_across_tasks(data, task_name, min_maps, max_maps, peaks_only=False):
    temp_data = np.zeros((64, 1))
    for a_task_name in task_name:
        opt_k_index = data[a_task_name]["opt_k_index"]
        temp = np.asarray(data[a_task_name]["maps_list"][opt_k_index]).T
        temp_data = np.concatenate((temp_data, temp), axis=1)
    temp = batch_microstate([temp_data[:, 1::], peaks_only, min_maps, max_maps])
    return {'cv_list': temp.cv_list, 'gev_list': temp.gev_list, 'maps_list': temp.maps_list,
            'opt_k': temp.opt_k, 'opt_k_index': int(temp.opt_k_index), 'min_maps': min_maps,
            'max_maps': max_maps}


def opt_microstate_across_subjects_across_tasks_across_runs(clean_data_fname, data_fname, data_fname_save, task_name, peaks_only=False):
    data_text = codecs.open(clean_data_fname + "\\" + data_fname, 'r', encoding='utf-8').read()
    data = json.loads(data_text)
    res = np.zeros((64, 1))
    for a_task_name in task_name:
        opt_k_index = data[a_task_name]["opt_k_index"]
        res = np.concatenate((res, np.asarray(data[a_task_name]["maps_list"][opt_k_index]).T), axis=1)

    temp = opt_microstate_tasks({'all_tasks': res[::, 1:]}, ['all_tasks'], [6], [6], peaks_only)
    json.dump(temp, codecs.open(clean_data_fname + "\\" + data_fname_save, 'w', encoding='utf-8'), separators=(',', ':'), sort_keys=True, indent=4)


def opt_microstate_subjects(clean_data_fname, subjects, data_fname, data_fname_save_dict, data_fname_save, task_name, min_maps, max_maps, across_tasks=False, peaks_only=False):
    for subject in subjects:
        print(subject)
        # data_text = codecs.open(clean_data_fname + "\\" + subject + data_fname, 'r', encoding='utf-8').read()
        data = load_data(clean_data_fname + "\\" + subject + data_fname)
        data = combine_task_data(data, task_name)
        res = opt_microstate_subjects_across_tasks(data, task_name, min_maps, max_maps, peaks_only) if across_tasks else opt_microstate_tasks(data, task_name, min_maps, max_maps, peaks_only)
        json.dump(res, codecs.open(data_fname_save_dict + "\\" + subject + data_fname_save, 'w', encoding='utf-8'), separators=(',', ':'), sort_keys=True)


def combine_task_data(data, combined_task):
    data_temp = OrderedDict()
    for task in combined_task:
        a_task = task.split("#")
        if len(a_task) > 1:
            temp = np.asarray(data[a_task[0]]['task_data'])
            for i in range(1, len(a_task)):
                temp = np.concatenate((temp, np.asarray(data[a_task[i]]['task_data'])),axis=1)
            data_temp[task] = temp.tolist()
        else:
            data_temp[task] = data[task]['task_data']
    return data_temp


def save_eegmpas(clean_data_fname, data_fname, task_name):
    data_text = codecs.open(clean_data_fname + "\\" + data_fname, 'r', encoding='utf-8').read()
    data = json.loads(data_text)
    microstate_map = []
    for a_task_name in task_name:
        microstate_map.append(np.asarray(data[a_task_name]["maps_list"][data[a_task_name]["opt_k_index"]]))
    plot_eegmaps(microstate_map, task_name)

def save_eegmaps_subjects(clean_data_fname, data_fname, task_name, subjects):
    for subject in subjects:
        save_eegmpas(clean_data_fname, subject+data_fname, task_name)


### updated by Mengting on March 16 2022 for CTA data processing
########## for optimal K (nb of eegmaps) for each task listing results from all subjects ################
## tested with success, function ready for use ########
def microstate_stat(clean_data_fname, subjects, data_fname, data_fname_save, task_name):
    res = OrderedDict()
    for subject in subjects:
        data_fname_save_dict = clean_data_fname + "\\" + subject + "\\" + 'microstates'
        data_text = codecs.open(data_fname_save_dict + "\\" + subject + data_fname, 'r', encoding='utf-8').read()
        data = json.loads(data_text)
        for a_task_name in task_name:
            if a_task_name not in res:
                res[a_task_name] = {'opt_k': [], 'min_maps': data[a_task_name]["min_maps"], 'max_maps': data[a_task_name]["max_maps"]}
            res[a_task_name]['opt_k'].append(data[a_task_name]["opt_k"])
    for a_task_name in task_name:
        res[a_task_name]['mean_opt_k'] = int(np.asarray(res[a_task_name]['opt_k']).mean())
        res[a_task_name]['mean_opt_k_index'] = res[a_task_name]['mean_opt_k'] - int(res[a_task_name]['min_maps'])
    json.dump(res, codecs.open(clean_data_fname + "\\" + 'microstates_stats' + "\\" + data_fname_save, 'w', encoding='utf-8'), separators=(',', ':'),sort_keys=True, indent=4)
    return res


def load_data(path):
    data_text = codecs.open(path, 'r', encoding='utf-8').read()
    data = json.loads(data_text)
    return data

def read_microstate_stat(data_fname):
    res = {}
    data_text = codecs.open(data_fname, 'r', encoding='utf-8').read()
    data = json.loads(data_text)
    for task_name, task_data in data.items():
        res[task_name] = {'mean_opt_k': task_data['mean_opt_k']}
    return res


def normalized_data(data):
    for i in range(data.shape[0]):
        for j in range(data.shape[1]):
            # data[i][j] = (data[i][j]-np.mean(data[i][j])) / np.std(data[i][j])
            data[i][j] = data[i][j]-np.mean(data[i][j])
    return data

def generalized_dissimilarity(data):
    grand_mean_across_subjects = np.mean(data, axis=0)
    grand_mean_across_subjects_across_conditions = np.mean(data, axis=(0, 1))

    # grand_mean_across_subjects = (grand_mean_across_subjects - np.mean(grand_mean_across_subjects, axis=0)) / np.std(grand_mean_across_subjects, axis=0)
    # grand_mean_across_subjects_across_conditions = (grand_mean_across_subjects_across_conditions - np.mean(grand_mean_across_subjects_across_conditions)) / np.std(grand_mean_across_subjects_across_conditions)

    # s = 0
    # for i in range(grand_mean_across_subjects.shape[0]):
    #     s += (stats.pearsonr(grand_mean_across_subjects[i], grand_mean_across_subjects_across_conditions)[0])
    residual = np.power(grand_mean_across_subjects - grand_mean_across_subjects_across_conditions, 2)
    s = np.sum(np.sqrt(np.mean(residual, axis=1)))
    return s


def format_data(data, condition, n_subject, n_ch, ith_class):
    n_condition = len(condition)
    data_temp = np.zeros((n_subject, n_condition, n_ch))
    for i in range(n_subject):
        for j in range(n_condition):
            data_temp[i,j] = data[condition[j]][i][ith_class]
    return data_temp

def shuffle_data(data, n_condition):
    for i in range(data.shape[0]):
        random_index = np.random.permutation(n_condition)
        data[i] = data[i][random_index]
    return data

def diss(v_ij, v_j):
    return np.sum(np.sqrt(np.mean(np.power(v_ij - v_j, 2),1)))

def diss_interaction(level_mean, grand_mean):
    residual = np.power(level_mean -grand_mean, 2)
    return np.sum(np.sqrt(np.mean(residual, 2)))


def tanova(data, condition, n_subjects, n_c, n_k, n_ch, n_times):
    data_concatenation = np.zeros((n_c, n_subjects, n_k, n_ch))
    for i, a_condition_name in enumerate(condition):
        temp = np.asarray(data[a_condition_name])
        # temp = np.asarray(data[a_condition_name]['maps'])  #changed by Mengting for analysis without re-order step
        temp = temp - temp.mean(axis=2).reshape(n_subjects, n_k, 1)
        data_concatenation[i] = temp
    condition_mean = data_concatenation.mean(axis=2)
    class_mean = data_concatenation.mean(axis=0)
    grand_mean = data_concatenation.mean(axis=(0, 1, 2))
    grand_mean = (grand_mean - grand_mean.mean()) / np.std(grand_mean)

    condition_mean_subject = condition_mean.mean(1)
    class_mean_subject = class_mean.mean(0)
    condition_mean_subject = (condition_mean_subject - condition_mean_subject.mean(1, keepdims=True)) / np.std(condition_mean_subject, axis=1,keepdims=True)
    class_mean_subject = (class_mean_subject - class_mean_subject.mean(1, keepdims=True)) / np.std(class_mean_subject, axis=1,keepdims=True)

    # condition_class_mean = data_concatenation - class_mean_subject - np.tile(condition_mean_subject, n_k).reshape(n_c,1, n_k, n_ch)
    condition_class_mean = data_concatenation - class_mean_subject
    condition_class_mean_subject = condition_class_mean.mean(1, keepdims=True)
    condition_class_mean_subject = (condition_class_mean_subject - condition_class_mean_subject.mean(axis=(0, 1), keepdims=True)) / np.std(condition_class_mean_subject, axis=(0, 1), keepdims=True)

    count = 0
    condition_s = []
    class_s = []
    condition_class_s = []
    while count != n_times:
        condition_s.append(diss(condition_mean_subject, grand_mean))
        class_s.append(diss(class_mean_subject, grand_mean))
        condition_class_s.append(diss_interaction(condition_class_mean_subject, grand_mean))

        for i in range(n_subjects):
            condition_order = np.random.permutation(n_c)
            class_order = np.random.permutation(n_k)
            condition_mean[:, i, :] = condition_mean[condition_order, i, :]
            class_mean[i, :, :] = class_mean[i, class_order, :]
            data_concatenation[:, i] = data_concatenation[:, i, class_order]

        condition_mean_subject = condition_mean.mean(1)
        class_mean_subject = class_mean.mean(0)
        condition_mean_subject = (condition_mean_subject - condition_mean_subject.mean(1, keepdims=True)) / np.std(condition_mean_subject, axis=1, keepdims=True)
        class_mean_subject = (class_mean_subject - class_mean_subject.mean(1, keepdims=True)) / np.std(class_mean_subject, axis=1, keepdims=True)

        # condition_class_mean = data_concatenation - class_mean_subject - np.tile(condition_mean_subject, n_k).reshape(n_c, 1, n_k, n_ch)
        condition_class_mean = data_concatenation - class_mean_subject
        condition_class_mean_subject = condition_class_mean.mean(1, keepdims=True)
        condition_class_mean_subject = (condition_class_mean_subject - condition_class_mean_subject.mean(axis=(0, 1),keepdims=True)) / np.std(condition_class_mean_subject, axis=(0, 1), keepdims=True)
        count += 1

    res = [len(np.where(condition_s >= condition_s[0])[0]) / n_times, len(np.where(class_s >= class_s[0])[0]) / n_times,
           len(np.where(condition_class_s >= condition_class_s[0])[0]) / n_times]

    return res





def permutation_test_2_conditions(data, task_name, n_subjects, n_times):
    p_list = []
    for comb in itertools.combinations(task_name, 2):
        print(comb)
        res = []
        temp_task_name = [comb[0], comb[1]]
        for i in range(6):
            data_format = format_data(data, temp_task_name, n_subjects, 64, i)
            data_normalized = normalized_data(data_format)
            count = 0
            temp = []
            while count != n_times:
                temp.append(generalized_dissimilarity(data_normalized))
                data_normalized = shuffle_data(data_normalized, len(temp_task_name))
                count += 1
            temp = np.asarray(temp)
            res.append(len(np.where(temp >= temp[0])[0]) / n_times)
        p_list.append(res)
    return p_list

def permutation_test_3_conditions(data, task_name, n_subjects, n_times):
    res = []
    for i in range(6):
        data_format = format_data(data, task_name, n_subjects, 64, i)
        data_normalized = normalized_data(data_format)
        count = 0
        temp = []
        while count != n_times:
            temp.append(generalized_dissimilarity(data_normalized))
            data_normalized = shuffle_data(data_normalized, len(task_name))
            count += 1
        temp = np.asarray(temp)

        res.append(len(np.where(temp >= temp[0])[0]) / n_times)
    return res

def outliers_index(data, outlierConstant=1.5):
    upper_quartile = np.percentile(data, 75)
    lower_quartile = np.percentile(data, 25)
    IQR = (upper_quartile - lower_quartile) * outlierConstant
    quartileSet = (lower_quartile - IQR, upper_quartile + IQR)
    index = np.where((data >= quartileSet[0]) & (data <= quartileSet[1]))[0]
    return index.tolist()

## updated by Mengting on April 2022 for CTA data analysis
def reorder_individual_eegmaps(global_maps, subjects, data_path, task, n_k):
    k_index = n_k -2  ### added by Mengting after removing one input parameter
    n_chs = 64

    subjects_fname = data_path # the updated file path with REST features
    for subject in subjects:
        # subjects_fname = data_dir + "\\"+ subject + "\\" + r'microstates'
        res = {}
        maps = []
        for a_task in task:
            maps.append(load_data(subjects_fname+"\\"+subject+"_microstate.json")[a_task]["maps_list"][k_index])
        m = MeanMicrostate(maps, n_k, n_chs, len(task))
        temp = m.reorder_microstates(np.asarray(maps), global_maps)
        for index, a_task in enumerate(task):
            res[a_task] = temp[index]
        json.dump(res, codecs.open(subjects_fname+"\\"+subject+"_two_microstate_reorder.json", 'w', encoding='utf-8'), separators=(',', ':'), sort_keys=True, indent=4)

## updated by Mengting on April 2022 for CTA data analysis
def reorder_baseline_individual_eegmaps(global_maps, subjects, data_path, task, n_k):
    k_index = n_k -2  ### added by Mengting after removing one input parameter
    n_chs = 64
    subjects_fname = data_path # the updated file path with REST features
    for subject in subjects:
        # subjects_fname = data_dir + "\\"+ subject + "\\" + r'microstates'
        res = {}
        maps = []
        for a_task in task:
            ### the naming of _baseline_microstate.json differs from Trial results, but the following steps will be the same
            maps.append(load_data(subjects_fname+"\\"+subject+"_baseline_microstate.json")[a_task]["maps_list"][k_index])
        m = MeanMicrostate(maps, n_k, n_chs, len(task))
        temp = m.reorder_microstates(np.asarray(maps), global_maps)
        for index, a_task in enumerate(task):
            res[a_task] = temp[index]
        json.dump(res, codecs.open(subjects_fname+"\\"+subject+"_two_microstate_reorder.json", 'w', encoding='utf-8'), separators=(',', ':'), sort_keys=True, indent=4)
### this function is different from the Microstate.fit_back defined in the Microstate class
###### updated by Mengting on March 16 2022 for CTA analysis
############## tested with success, ready for use ###########
def fit_back(data_dir, maps_fname, data_fname_dict, subjects, tasks, is_global_maps, is_instantaneous=False):
    for subject in subjects:
        print(subject)
        temp_res = []
        # res = {}
        subject_path = pjoin(data_dir, subject)
        set_path = pjoin(subject_path, r'clean_data_matlab\data')
        # data_fname_dict = pjoin(subject_path, r'microstates')
        res = OrderedDict()
        data = OrderedDict()
        for q, a_task in enumerate(tasks):
            # min_maps_lis = [1 for i in range(len(path_name))]
            # max_maps_lis = [10 for i in range(len(path_name))]
            trial_path = pjoin(set_path, path_name[q])
            set_name = pjoin(trial_path, lis_name[q])
            EEG_epochs= mne.io.read_epochs_eeglab(set_name)

            # deal with the channel naming issue when adding .loc information
            mapping = {'Afz': 'AFz'}   # an error exists in the generated epoch data from N-Back test
            EEG_epochs.rename_channels(mapping, allow_duplicates=True, verbose=None)
            names_channel = EEG_epochs.ch_names
            n_chs = len(names_channel)

            EEG_epochs = EEG_epochs.set_montage('biosemi64')
            nb_epochs = EEG_epochs.selection.size
            # n_epo[q] = nb_epochs
            nb_epochs = 4  ## for testing purpose

            # data_test = EEG_epochs.get_data()
            data_indtrial = []   # concatenate the data for one trial (from multiple epochs)
            for i in range(nb_epochs):
                temp_epoch = EEG_epochs[i]  # get one epoch at a time
                temp_evoked = temp_epoch.average()  # this data structure corresponds to that used in preprocessed_nback.py
                temp_key = list(temp_epoch.event_id)[0]
                temp_data = temp_epoch.get_data()[0]
                if i ==0:
                    data_indtrial = temp_data
                else:
                    data_indtrial = np.concatenate((data_indtrial,temp_data), axis=1)
            data[a_task] = data_indtrial
        if is_global_maps:
            # # template = load_data(maps_fname)
            maps_data_text = codecs.open(maps_fname, 'r', encoding='utf-8').read()
            template = np.asarray(json.loads(maps_data_text))
            # maps_data_fname_save_dict = data_dir + "\\" + 'microstates_subjects'
            # maps_data_fname = 'reorder_microstate_across_runs_across_participants_across_conditions.json'  #use the reordered global maps
            # maps_data_text = codecs.open(maps_data_fname_save_dict + "\\" + maps_data_fname, 'r', encoding='utf-8').read()
            # # template = np.asarray(json.loads(maps_data_text)["maps"])
            # template = np.asarray(json.loads(maps_data_text))
            for a_task in tasks:
                data_temp = zero_mean(np.asarray(data[a_task]).T, 1)
                vvv = Microstate.fit_back(data_temp, template, instantaneous_eeg=is_instantaneous)
                temp_res.append(Microstate.fit_back(data_temp, template, instantaneous_eeg=is_instantaneous))
                # temp_res.append(pool.apply_async(batch_fit_back, ([data_temp, np.asarray(template), is_instantaneous],)))
            for i, a_task in enumerate(tasks):
                res[a_task] = temp_res[i].tolist()
            json.dump(res, codecs.open(data_fname_dict + "\\" + subject + "_two_microstate_labels.json", 'w', encoding='utf-8'), separators=(',', ':'))
        else:
            for a_task in tasks:
                template = load_data(maps_fname + "\\" +subject +"_microstate.json")[a_task]
                index = template['opt_k_index']
                template_map = template['maps_list'][index]
                Microstate.fit_back(zero_mean(np.asarray(data[a_task]).T, 1), np.asarray(template_map), instantaneous_eeg=is_instantaneous)



def batch_fit_back(para):
    data = para[0]
    template = para[1]
    is_instantaneous = para[2]
    return Microstate.fit_back(data, template, instantaneous_eeg=is_instantaneous).tolist()


# def fit_back(maps_fname, subjects, subjects_fname, subjects_sfname, task, is_global_maps):
#     for subject in subjects:
#         print(subject)
#         res = {}
#         data = load_data(subjects_fname + "\\" + subject + '.json')
#         data = combine_task_data(data, task)
#
#         if is_global_maps:
#             template = load_data(maps_fname)['maps']
#             for a_task in task:
#                 data_temp = zero_mean(np.asarray(data[a_task]).T, 1)
#                 res[a_task] = Microstate.fit_back(data_temp, np.asarray(template), instantaneous_eeg=False).tolist()
#         else:
#             template = load_data(maps_fname + "\\" + subject +"_1_30_microstate_reorder.json")
#             for a_task in task:
#                 data_temp = zero_mean(np.asarray(data[a_task]).T, 1)
#                 res[a_task] = Microstate.fit_back(data_temp, np.asarray(template[a_task]),instantaneous_eeg=True).tolist()
#         json.dump(res, codecs.open(subjects_sfname + "\\" + subject + "_microstate_labels.json", 'w',
#                                    encoding='utf-8'), separators=(',', ':'), sort_keys=True)

################ updated on March14 2022 by Mengting Zhao for CTA analysis for pilot training process###################
######### the function is ready for use
def eegmaps_individual_run(data_dir, subjects, data_fname_save, task_name, min_maps, max_maps, peaks_only=True):
    # the content of this function has been tested in the loop for each trial
    # data_fname_save="_microstate.json"

    for subject in subjects:
        print(subject)
        # data = load_data(clean_data_fname + "\\" + subject + data_fname)  #from Wenjun's code
        subject_path = pjoin(data_dir, subject)
        # data_fname_save_dict = pjoin(subject_path, r'microstates')
        res = OrderedDict()
        n_epo = np.zeros(nb_trials)
        for q in range(nb_trials):
            min_maps_lis = [1 for i in range(len(path_name))]
            max_maps_lis = [10 for i in range(len(path_name))]
            trial_path = subject_path + "\\" + r'clean_data_matlab' + "\\" + r'data' + "\\" + path_name[q]
            set_name = pjoin(trial_path, lis_name[q])
            EEG_epochs= mne.io.read_epochs_eeglab(set_name)

            # deal with the channel naming issue when adding .loc information
            mapping = {'Afz': 'AFz'}   # an error exists in the generated epoch data from N-Back test
            EEG_epochs.rename_channels(mapping, allow_duplicates=True, verbose=None)
            names_channel = EEG_epochs.ch_names
            n_chs = len(names_channel)

            EEG_epochs = EEG_epochs.set_montage('biosemi64')
            # # get the IDs of the events in the dataset
            # ids_event = EEG_epochs.event_id
            nb_epochs = EEG_epochs.selection.size
            n_epo[q] = nb_epochs

            data_test = EEG_epochs.get_data()
            temp_res = []   # concatenate the data for one trial (from multiple epochs)
            # nb_epochs = 5   ## for testing purpose
            for i in range(nb_epochs):
                temp_epoch = EEG_epochs[i]  # get one epoch at a time
                temp_evoked = temp_epoch.average()  # this data structure corresponds to that used in preprocessed_nback.py
                temp_key = list(temp_epoch.event_id)[0]
                temp_data = temp_epoch.get_data()[0]
                if i ==0:
                    temp_res = temp_data
                else:
                    temp_res = np.concatenate((temp_res,temp_data), axis=1)

            ## generate an instance of Microstate for each-trial data
            #######  correspond to eegmaps_individual_run() function
            micro_test = Microstate(temp_res)
            micro_test.gfp_peaks()  # compute the GFP peaks, and send the EEG signals at the FFP peaks for Microstate analysi
            # # temp_test.kmeans_modified()   #indluced in opt_microstate with method='kmeans_modified'
            micro_test.opt_microstate(min_maps, max_maps, n_std=3, peaks_only=peaks_only, method='kmeans_modified')  #the global microstate classes are determined at the end of this function

            # the fitting-back process (fit_back() function) is called in the microstates_parameters() function for occurrence, coverage, and duration computation
            ### testing Microstate.opt_microstates() function
            res[task_name[q]] = {'cv_list': micro_test.cv_list, 'gev_list': micro_test.gev_list, 'maps_list': micro_test.maps_list,
                                 'opt_k': micro_test.opt_k, 'opt_k_index': int(micro_test.opt_k_index), 'min_maps': min_maps_lis[q],
                                 'max_maps': max_maps_lis[q]}
        json.dump(res, codecs.open(subject_path + "\\" + r'microstates' + "\\" + subject + data_fname_save, 'w', encoding='utf-8'), separators=(',', ':'))


## updated on March14 2022 by Mengting Zhao for CTA analysis for pilot training process
### the function is ready for use ########
## the content has been tested  in main() function with success
def eegmaps_across_runs(data_dir, subjects, task_name, condition_name, n_k, n_k_index, n_chs):
    ## eegmaps_across_runs(clean_data_fname, subjects, data_fname, data_fname_save_dict, data_fname_save, task_name, condition_name, n_k, n_k_index, n_ch):
    # clean_data_fname, subjects, '_microstate.json', data_fname_save_dict, '_microstate_across_runs.json', tasks, conditions, 7, 6, 63
    conditions = {}
    res = OrderedDict()
    for condition in condition_name:
        conditions[condition] = []
        for task in task_name:
            if task.startswith(condition):
                conditions[condition].append(task)
    for subject in subjects:
        data_fname_save_dict = data_dir + "\\" + subject + "\\" + 'microstates'
        # data_fname_save_dict = pjoin(subject_path, r'microstates')
        print(subject)
        ## the input is the result from individual_run, which is P01_microstate.json for example
        data_text = codecs.open(data_fname_save_dict + "\\" + subject + '_microstate.json', 'r', encoding='utf-8').read()
        # data = OrderedDict()
        data = json.loads(data_text)
        condition_res = []
        # pool = Pool(len(condition_name))
        res = OrderedDict()
        for condition in condition_name:
            maps = [data[a_task_name]['maps_list'][n_k_index] for a_task_name in conditions[condition]]
            # condition_res.append(pool.apply_async(batch_mean_microstate, ([maps, n_k, n_chs, len(conditions[condition])],)))
            vvv = MeanMicrostate(maps, n_k, n_chs, len(conditions[condition]))
            eegmaps, label, mean_similarity, std_similarity = vvv.mean_microstates()
            condition_res.append({"maps": eegmaps, "label": label, "mean_similarity": mean_similarity, "std_similarity": std_similarity})
            res[condition] = {"maps": eegmaps.tolist(), "label": label, "mean_similarity": mean_similarity, "std_similarity": std_similarity}
            print(9)
        print(000)
        json.dump(res, codecs.open(data_fname_save_dict + "\\" + subject + '_microstate_across_runs.json', 'w', encoding='utf-8'), separators=(',', ':'), sort_keys=True)




if __name__ == '__main__':
    data_dir = r'F:\FlightSim_experiment_2021\Concordia_preprocessed'
    # for processing pre-processed segmentation data for FlighSim CTA data
    ########### for testing the Meanmicrostate class, let's define the conditions to be studied:
    ############ 0: the two resting sessions: eyes-open and eyes-closed
    ########### 1. the first 1-7 trials (training)
    ########### 2. trial 8-14  (7 trials out of the 8 trials included in PracticeA)
    ########## 3. trial 16-22  (7 trials in PracticeB)
    subjects = ['P01', 'P02', 'P03', 'P04', 'P05', 'P06', 'P07', 'P08', 'P09', 'P10', 'P11', 'P12',
                'P13', 'P14', 'P15', 'P16', 'P17', 'P18', 'P19', 'P20', 'P21', 'P22', 'P23', 'P24']

    condition_name = ['training', 'practiceA', 'practiceB']
    n_conditions = len(condition_name)
    ## but the expected results actually contain 7 conditions ['rest', 'training', 'practiceA', 'practiceB', 'B_training', 'B_practiceA', 'B_practiceB']

    task_name = ['rest_1', 'rest_2',
                 'training_1', 'training_2', 'training_3', 'training_4', 'training_5', 'training_6', 'training_7',
                 'practiceA_1', 'practiceA_2', 'practiceA_3', 'practiceA_4', 'practiceA_5', 'practiceA_6', 'practiceA_7', 'practiceA_8',
                 'practiceB_1', 'practiceB_2', 'practiceB_3', 'practiceB_4', 'practiceB_5', 'practiceB_6', 'practiceB_7']
    # nb_trials = len(task_name)

    ##########################  across-runs-across-subjects-across-conditions (global maps) ####################################
    # ######## successful testing ###################################
    # n_k = 7  # using 7 microstates
    # n_k_index = n_k - 2
    # n_chs = 64
    #
    # # ##### for computing on Mengting's computer
    # clean_data_dict_1 = r'F:\FlightSim_experiment_2021\Concordia_preprocessed\microstates_subjects'
    # fname_1 = r'_microstate_across_runs_across_participants.json'
    # clean_data_dict_2 = r'F:\FlightSim_experiment_2021\Concordia_preprocessed\baseline_microstate_analysis\microstates_subjects'
    # fname_2 = r'_baseline_microstate_across_runs_across_participants.json'
    #
    #
    # data_fname_save_dict = clean_data_dict_1
    # two_eegmaps_across_conditions(clean_data_dict_1, fname_1, clean_data_dict_2, fname_2, data_fname_save_dict, 'two_microstate_across_runs_across_participants_across_conditions.json', condition_name, n_k, n_chs)
    #
    # print('cross-conditions testing')


    ################### Step N: for labeling the original data with the reordered global eegmaps #################################
    ### to be computed on Mengting's laptop, as original EEG data are needed
    ### the results (_labels.json) files generated from this process are used for generating Surrogate data in microstate_sequence_analysis_zhao.py
    #### test the fit_back() function, ???? what's the difference with the _seq.mat process renerated in "microstate_to_sequence.py"
        data_dir = r'F:\FlightSim_experiment_2021\Concordia_preprocessed'
        global_eegmaps_fname = data_dir + "\\" + "microstates_subjects" + "\\" + "reorder_two_microstate_across_runs_across_participants_across_conditions.json"

        # ### for Trial computation
        # path_name = ['eyes_open', 'eyes_closed',
        #              'trial_1', 'trial_2', 'trial_3', 'trial_4', 'trial_5', 'trial_6', 'trial_7',
        #              'trial_8', 'trial_9', 'trial_10', 'trial_11', 'trial_12', 'trial_13', 'trial_14', 'trial_15',
        #              'trial_16', 'trial_17', 'trial_18', 'trial_19', 'trial_20', 'trial_21', 'trial_22']
        # lis_name = ['eyes_open.set', 'eyes_closed.set',
        #         'trial_1.set', 'trial_2.set', 'trial_3.set', 'trial_4.set', 'trial_5.set', 'trial_6.set', 'trial_7.set',
        #         'trial_8.set', 'trial_9.set', 'trial_10.set', 'trial_11.set', 'trial_12.set', 'trial_13.set', 'trial_14.set', 'trial_15.set',
        #         'trial_16.set', 'trial_17.set', 'trial_18.set', 'trial_19.set', 'trial_20.set', 'trial_21.set', 'trial_22.set']
        # data_fname_dict = r'F:\FlightSim_experiment_2021\Concordia_preprocessed\microstates_subjects\trial_individual'
        # fit_back(data_dir, global_eegmaps_fname, data_fname_dict, subjects, task_name, True)
        # print('Results computed on my laptop')

        ### for Baseline computation
        lis_name = ['eyes_open.set', 'eyes_closed.set','baseline_1.set', 'baseline_2.set', 'baseline_3.set', 'baseline_4.set', 'baseline_5.set', 'baseline_6.set',
                    'baseline_7.set', 'baseline_8.set', 'baseline_9.set', 'baseline_10.set', 'baseline_11.set', 'baseline_12.set',
                    'baseline_13.set', 'baseline_14.set', 'baseline_15.set', 'baseline_16.set', 'baseline_17.set', 'baseline_18.set',
                    'baseline_19.set', 'baseline_20.set', 'baseline_21.set', 'baseline_22.set']

        path_name = ['eyes_open', 'eyes_closed','baseline_1', 'baseline_2', 'baseline_3', 'baseline_4', 'baseline_5', 'baseline_6',
                     'baseline_7', 'baseline_8', 'baseline_9', 'baseline_10', 'baseline_11', 'baseline_12',
                     'baseline_13', 'baseline_14', 'baseline_15', 'baseline_16', 'baseline_17', 'baseline_18',
                     'baseline_19', 'baseline_20', 'baseline_21', 'baseline_22']
        data_fname_dict = r'F:\FlightSim_experiment_2021\Concordia_preprocessed\baseline_microstate_analysis\microstates_subjects\baseline_individual'
        fit_back(data_dir, global_eegmaps_fname, data_fname_dict, subjects, task_name, True)
        print('Results computed on my laptop')


